@@ -1,6 +1,6 @@
 /*
  * madplay - MPEG audio decoder and player
- * Copyright (C) 2000-2003 Robert Leslie
+ * Copyright (C) 2000-2004 Robert Leslie
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -16,7 +16,7 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  *
- * $Id: madplay.c,v 1.82 2003/06/05 03:56:04 rob Exp $
+ * $Id: madplay.c,v 1.85 2004/02/17 02:26:43 rob Exp $
  */
 
 # ifdef HAVE_CONFIG_H
@@ -79,10 +79,13 @@ struct option const options[] = {
   { "mono",		no_argument,		0,		 'm' },
   { "no-dither",	no_argument,		0,		 'd' },
   { "output",		required_argument,	0,		 'o' },
+  { "pre-amp",		required_argument,	0,		 'a' },
   { "quiet",		no_argument,		0,		 'q' },
   { "repeat",		optional_argument,	0,		 'r' },
+  { "replay-gain",	optional_argument,	0,		 'G' },
   { "right",		no_argument,		0,		 '2' },
   { "sample-rate",	required_argument,	0,		 'R' },
+  { "show-tags-only",	no_argument,		0,		 'T' },
   { "shuffle",		no_argument,		0,		 'z' },
   { "start",		required_argument,	0,		 's' },
   { "stereo",		no_argument,		0,		 'S' },
@@ -108,7 +111,7 @@ struct option const options[] = {
 
 char const *argv0;
 
-# define EPUTS(str)	fputs(str, stderr)
+# define EPUTS(str)	fputs(str, stream)
 
 /*
  * NAME:	show_usage()
@@ -117,10 +120,12 @@ char const *argv0;
 static
 void show_usage(int verbose)
 {
-  fprintf(stderr, _("Usage: %s [OPTIONS] FILE [...]\n"), argv0);
+  FILE *stream = verbose ? stdout : stderr;
+
+  fprintf(stream, _("Usage: %s [OPTIONS] FILE [...]\n"), argv0);
 
   if (!verbose) {
-    fprintf(stderr, _("Try `%s --help' for more information.\n"), argv0);
+    fprintf(stream, _("Try `%s --help' for more information.\n"), argv0);
     return;
   }
 
@@ -150,11 +155,11 @@ void show_usage(int verbose)
 	  "request HERTZ samples per second\n"));
   EPUTS(_("  -d, --no-dither              "
 	  "do not dither output PCM samples\n"));
-  fprintf(stderr,
+  fprintf(stream,
 	_("      --fade-in[=DURATION]     "
 	  "fade-in songs over DURATION (default %s)\n"), FADE_DEFAULT);
 # if 0
-  fprintf(stderr,
+  fprintf(stream,
 	_("      --fade-out[=DURATION]    "
 	  "fade-out songs over DURATION (default %s)\n"), FADE_DEFAULT);
   EPUTS(_("  -g, --gap=DURATION           set inter-song gap to DURATION\n"));
@@ -166,6 +171,9 @@ void show_usage(int verbose)
   EPUTS(_("  -a, --amplify=DECIBELS       amplify signal by DECIBELS (+)\n"));
   EPUTS(_("  -A, --adjust-volume=DECIBELS "
 	  "override per-file volume adjustments\n"));
+  EPUTS(_("  -G, --replay-gain[=PROFILE]  "
+	  "enable Replay Gain volume adjustments using\n"
+	  "                                 PROFILE (radio, audiophile)\n"));
 
   EPUTS(_("\nChannel selection:\n"));
   EPUTS(_("  -1, --left                   "
@@ -197,16 +205,19 @@ void show_usage(int verbose)
 # endif
 
   EPUTS(_("\nMiscellaneous:\n"));
+  EPUTS(_("  -T, --show-tags-only         "
+	  "show ID3/encoder tags only (do not decode)\n"));
   EPUTS(_("  -V, --version                display version number and exit\n"));
   EPUTS(_("      --license                "
 	  "show copyright/license message and exit\n"));
   EPUTS(_("  -h, --help                   display this help and exit\n"));
 
   EPUTS(_("\nSupported output formats:\n"));
-  EPUTS(_("  cdda    CD audio, 16-bit 44100 Hz stereo PCM (*.cdr, *.cda)\n"));
+  EPUTS(_("  cdda    "
+	  "CD audio, 16-bit big-endian 44100 Hz stereo PCM (*.cdr, *.cda)\n"));
   EPUTS(_("  aiff    Audio IFF, [16-bit] PCM (*.aif, *.aiff)\n"));
   EPUTS(_("  wave    Microsoft RIFF/WAVE, [16-bit] PCM (*.wav)\n"));
-  fprintf(stderr,
+  fprintf(stream,
 	_("  snd     Sun/NeXT audio, 8-bit ISDN %s (*.au, *.snd)\n"),
 	  _("mu-law"));
   EPUTS(_("  raw     binary [16-bit] host-endian linear PCM\n"));
@@ -483,18 +494,18 @@ static
 void get_options(int argc, char *argv[], struct player *player)
 {
   int opt, index;
-  int ttyset = 0;
+  int ttyset = 0, preamp = 0;
 
   while ((opt = getopt_long(argc, argv,
 			    "vqQ"		/* verbosity options */
 			    "i"			/* decoding options */
-			    "o:b:R:da:A:"	/* audio output options */
+			    "o:b:R:da:A:G::"	/* audio output options */
 # if 0
 			    "g:x"
 # endif
 			    "12mS"		/* channel selection options */
 			    "s:t:zr::"		/* playback options */
-			    "Vh",		/* miscellaneous options */
+			    "TVh",		/* miscellaneous options */
 			    options, &index)) != -1) {
     switch (opt) {
     case 0:
@@ -507,6 +518,7 @@ void get_options(int argc, char *argv[], struct player *player)
 
     case 'a':
       player->output.attamp_db = get_decibels(optarg);
+      preamp = 1;
       break;
 
     case 'A':
@@ -556,6 +568,26 @@ void get_options(int argc, char *argv[], struct player *player)
       break;
 # endif
 
+    case 'G':
+      player->output.replay_gain |= RGAIN_ENABLED;
+
+      if (optarg) {
+	size_t len;
+
+	len = strlen(optarg);
+
+	if (len && strncmp(optarg, "audiophile", len) == 0)
+	  player->output.replay_gain |= RGAIN_AUDIOPHILE;
+	else if (len && strncmp(optarg, "radio", len) == 0)
+	  ;
+	else
+	  die(_("invalid Replay Gain argument \"%s\""), optarg);
+      }
+
+      if (!preamp)
+	player->output.attamp_db = +6;  /* default pre-amp */
+      break;
+
     case 'h':
       show_usage(1);
       exit(0);
@@ -659,6 +691,10 @@ void get_options(int argc, char *argv[], struct player *player)
       }
       break;
 
+    case 'T':
+      player->options |= PLAYER_OPTION_SHOWTAGSONLY;
+      break;
+
     case 'v':
       player->verbosity = +1;
       break;
@@ -744,12 +780,26 @@ int main(int argc, char *argv[])
       warn(_("cross-fade ignored without negative gap"));
   }
 
+  if (player.output.replay_gain & RGAIN_ENABLED) {
+    if (player.options & PLAYER_OPTION_IGNOREVOLADJ)
+      warn(_("volume adjustment ignored with Replay Gain enabled"));
+    else
+      player.options |= PLAYER_OPTION_IGNOREVOLADJ;
+  }
+
+  if ((player.options & PLAYER_OPTION_SHOWTAGSONLY) &&
+      player.repeat != 1) {
+    warn(_("ignoring repeat"));
+    player.repeat = 1;
+  }
+
   /* make stop time absolute */
   if (player.options & PLAYER_OPTION_TIMED)
     mad_timer_add(&player.global_stop, player.global_start);
 
   /* get default audio output module */
-  if (player.output.command == 0)
+  if (player.output.command == 0 &&
+      !(player.options & PLAYER_OPTION_SHOWTAGSONLY))
     player.output.command = audio_output(0);
 
 # if defined(EXPERIMENTAL)
